#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <GL/glut.h>

#include "argument.h"
#include "socket.h"
#include "lib.h"
#include "Obj.h"
#include "Saucer.h"
#include "Block.h"
#include "Log.h"

static int saucer_num = 4;
static int track_num = 4;
static int sector_num = 4;

struct {
  double angle_dir;
  double angle_up;
  double distance;
  vector_t position;
  vector_t direction;
  vector_t updir;
} camera;

static int gl_camera()
{
  vector_t p;
  double rad;

  vector_set(&camera.position, camera.distance, 0.0, 0.0);

  rad = ((2.0 * M_PI) / 360.0) * camera.angle_up;
  p = camera.position;
  camera.position.x = cos(rad) * p.x - sin(rad) * p.z;
  camera.position.z = sin(rad) * p.x + cos(rad) * p.z;

  rad = ((2.0 * M_PI) / 360.0) * camera.angle_dir;
  p = camera.position;
  camera.position.x = cos(rad) * p.x - sin(rad) * p.y;
  camera.position.y = sin(rad) * p.x + cos(rad) * p.y;

  vector_mul(&camera.direction, &camera.position, -1.0);
  vector_set(&camera.updir, 0.0, 0.0, 1.0);

  return 0;
}

static int gl_init()
{
  GLfloat ambient[]  = { 0.4, 0.4, 0.4, 1.0 };
  GLfloat diffuse[]  = { 0.7, 0.7, 0.7, 1.0 };
  GLfloat specular[] = { 0.1, 0.1, 0.1, 1.0 };

  glClearColor(0.0, 0.0, 0.0, 1.0);
  glEnable(GL_DEPTH_TEST);

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glEnable(GL_LIGHTING);

  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

  glLightfv(GL_LIGHT0, GL_AMBIENT,  ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, specular);

  return 0;
}

static int gl_light()
{
  GLfloat position[] = { 0.0, 0.0, 2.0, 1.0 };
  glLightfv(GL_LIGHT0, GL_POSITION, position);
  glEnable(GL_LIGHT0);
  return 0;
}

static ObjList saucers;
static ObjList users;
static ObjList genres;

static int saucers_pattern(int pattern, int n, vector_t *points, vector_t *angles)
{
  int i;
  vector_t p;
  vector_t angle;

  for (i = 0; i < n; i++) {
    vector_zero(&p);
    vector_zero(&angle);
    switch (pattern % 26) {
    case 0:
      p.x = cos(2.0 * M_PI * i * 0.3) * 0.8;
      p.y = sin(2.0 * M_PI * i * 0.3) * 0.8;
      p.z = (double)(i + 1) / n * 2.0 - 1.0;
      break;

    case 1:
      p.x = ((rand() % 201) - 100) * 0.01;
      p.y = ((rand() % 201) - 100) * 0.01;
      p.z = ((rand() % 201) - 100) * 0.01;
      break;

    case 2:
      p.x = ((rand() % 201) - 100) * 0.01;
      p.y = ((rand() % 201) - 100) * 0.01;
      p.z = ((rand() % 201) - 100) * 0.01;
      angle.x = rand() % 360;
      angle.y = rand() % 360;
      angle.z = rand() % 360;
      break;

    case 3:
      p.x = cos(2.0 * M_PI * ((double)i / n)) * 1.5;
      p.y = sin(2.0 * M_PI * ((double)i / n)) * 1.5;
      break;

    case 4:
      p.x = cos(2.0 * M_PI * ((double)i / n)) * 1.5;
      p.y = sin(2.0 * M_PI * ((double)i / n)) * 1.5;
      p.z = (i % 2) * 0.2;
      break;

    case 5:
      p.x = cos(2.0 * M_PI * ((double)i / n)) * 1.5;
      p.y = sin(2.0 * M_PI * ((double)i / n)) * 1.5;
      p.z = (i % 2) * 0.2;
      angle.x = (1 - (i % 2)) * 180;
      break;

    case 6:
      p.x = cos(2.0 * M_PI * ((double)i / n)) * 1.5;
      p.y = sin(2.0 * M_PI * ((double)i / n)) * 1.5;
      p.z = (i % 2) * 0.2;
      angle.y = 90.0;
      angle.z = 360.0 * ((double)i / n);
      break;

    case 7:
      p.x = cos(2.0 * M_PI * ((double)i / n)) * 1.2;
      p.y = sin(2.0 * M_PI * ((double)i / n)) * 1.2;
      p.z = (double)i / n * 2.0 - 1.0;
      break;

    case 8:
      p.x = cos(2.0 * M_PI * ((double)(i * 2.5) / n)) * 1.2;
      p.y = sin(2.0 * M_PI * ((double)(i * 2.5) / n)) * 1.2;
      p.z = (double)i / n * 2.0 - 1.0;
      break;

    case 9:
      p.x = cos(2.0 * M_PI * ((double)i / n)) * 1.2;
      p.y = sin(2.0 * M_PI * ((double)i / n)) * 1.2;
      p.z = ((i % 2) * 2 - 1) * 0.4;
      break;

    case 10:
      p.x = cos(2.0 * M_PI * ((double)i / n)) * 1.2;
      p.y = sin(2.0 * M_PI * ((double)i / n)) * 1.2;
      p.z = ((i % 2) * 2 - 1) * 0.4 + ((i / 2) % 2) * 0.2;
      break;

    case 11:
      p.x = cos(2.0 * M_PI * ((double)(i / 4) / (n / 4))) * 1.2;
      p.y = sin(2.0 * M_PI * ((double)(i / 4) / (n / 4))) * 1.2;
      p.z = ((i % 4) - 1.5) * 0.5;
      break;

    case 12:
      p.x = cos(2.0 * M_PI * (((double)(i / 4) + 0.5 * (i % 2)) / (n / 4))) * 1.2;
      p.y = sin(2.0 * M_PI * (((double)(i / 4) + 0.5 * (i % 2)) / (n / 4))) * 1.2;
      p.z = ((i % 4) - 1.5) * 0.5;
      break;

    case 13:
      p.x = cos(2.0 * M_PI * (((double)(i / 4) + 0.5 * (i % 2)) / (n / 4))) * 1.2;
      p.y = sin(2.0 * M_PI * (((double)(i / 4) + 0.5 * (i % 2)) / (n / 4))) * 1.2;
      p.z = ((i % 4) - 1.5) * 0.5;
      angle.y = 90.0;
      angle.z = 360.0 * (((double)(i / 4) + 0.5 * (i % 2)) / (n / 4));
      break;

    case 14:
      p.z = (double)i / n * 2.0 - 1.0;
      break;

    case 15:
      if (i < 3) {
	p.x = cos(2.0 * M_PI * ((double)i / 3)) * 0.5;
	p.y = sin(2.0 * M_PI * ((double)i / 3)) * 0.5;
      } else if (i < 9) {
	p.x = cos(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.2;
	p.y = sin(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.2;
      } else {
	p.x = cos(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.9;
	p.y = sin(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.9;
      }
      break;

    case 16:
      if (i < 3) {
	p.x = cos(2.0 * M_PI * ((double)i / 3)) * 0.5;
	p.y = sin(2.0 * M_PI * ((double)i / 3)) * 0.5;
      } else if (i < 9) {
	p.x = cos(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.2;
	p.y = sin(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.2;
      } else {
	p.x = cos(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.9;
	p.y = sin(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.9;
      }
      p.z = (i % 2) * 0.2;
      break;

    case 17:
      if (i < 3) {
	p.x = cos(2.0 * M_PI * ((double)i / 3)) * 0.5;
	p.y = sin(2.0 * M_PI * ((double)i / 3)) * 0.5;
      } else if (i < 9) {
	p.x = cos(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.2;
	p.y = sin(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.2;
      } else {
	p.x = cos(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.9;
	p.y = sin(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.9;
      }
      p.z = ((rand() % 5) - 2) * 0.1;
      break;

    case 18:
      if (i < 3) {
	p.x = cos(2.0 * M_PI * ((double)i / 3)) * 0.7;
	p.y = sin(2.0 * M_PI * ((double)i / 3)) * 0.7;
	p.z = 0.7;
      } else if (i < 9) {
	p.x = cos(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.0;
	p.y = sin(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.0;
	p.z = 0.2;
      } else {
	p.x = cos(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.5;
	p.y = sin(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.5;
	p.z = -0.3;
      }
      break;

    case 19:
      if (i < 3) {
	p.x = cos(2.0 * M_PI * ((double)i / 3)) * 0.7;
	p.y = sin(2.0 * M_PI * ((double)i / 3)) * 0.7;
	p.z = -0.5;
      } else if (i < 9) {
	p.x = cos(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.0;
	p.y = sin(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.0;
	p.z = 0.0;
      } else {
	p.x = cos(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.5;
	p.y = sin(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.5;
	p.z = 0.5;
      }
      break;

    case 20:
      p.x = ((i % 4) - 1.5) * 0.8;
      p.z = ((i / 4) - (((n + 3) / 4) - 1) * 0.5) * 0.4;
      break;

    case 21:
      p.x = ((i % 4) - 1.5) * 0.8;
      p.z = ((i / 4) - (((n + 3) / 4) - 1) * 0.5) * 0.4 + (i % 2) * 0.1;
      break;

    case 22:
      p.x = cos(2.0 * M_PI * ((double)i / n)) * 1.2;
      p.z = sin(2.0 * M_PI * ((double)i / n)) * 1.2;
      break;

    case 23:
      if ((i % 2) == 0) {
	p.x = cos(2.0 * M_PI * ((double)(i + 1.0) / n)) * 1.2;
	p.z = sin(2.0 * M_PI * ((double)(i + 1.0) / n)) * 1.2;
      } else {
	p.y = cos(2.0 * M_PI * ((double)(i + 0.0) / n)) * 1.2;
	p.z = sin(2.0 * M_PI * ((double)(i + 0.0) / n)) * 1.2;
      }
      break;

    case 24:
      if ((i % 2) == 0) {
	p.x = cos(2.0 * M_PI * ((double)(i + 1.0) / n)) * 1.2;
	p.z = sin(2.0 * M_PI * ((double)(i + 1.0) / n)) * 1.2;
	angle.y = -(360.0 * ((double)(i + 1.0) / n) - 90.0);
      } else {
	p.y = cos(2.0 * M_PI * ((double)(i + 0.0) / n)) * 1.2;
	p.z = sin(2.0 * M_PI * ((double)(i + 0.0) / n)) * 1.2;
	angle.x = 360.0 * ((double)(i + 0.0) / n) - 90.0;
      }
      break;

    case 25:
    default:
      break;
    }

    points[i] = p;
    angles[i] = angle;
  }

  return 0;
}

static int set_saucers(int n)
{
  Saucer saucer;
  ObjListData obj;
  static int number = 0;
  int i = 0, len;
  vector_t *points;
  vector_t *angles;

  number += n;

  len = ObjList_GetLength(saucers);
  points = malloc(sizeof(*points) * len);
  angles = malloc(sizeof(*angles) * len);
  saucers_pattern(number, len, points, angles);

  for (obj = ObjList_GetStart(saucers);
       !ObjList_IsEndEdge(saucers, obj);
       obj = ObjListData_GetNext(obj)) {
    saucer = (Saucer)ObjListData_GetObj(obj);
    Saucer_Move(saucer, 10, points[i].x, points[i].y, points[i].z);
    Saucer_Rotate(saucer, 10, angles[i].x, angles[i].y, angles[i].z);
    i++;
  }

  free(points);
  free(angles);

  return 0;
}

static int make_saucers()
{
  Saucer saucer;
  int i;

  saucers = ObjList_Create();
  users = ObjList_Create();
  genres = ObjList_Create();

  for (i = 0; i < saucer_num; i++) {
    saucer = Saucer_Create(0.0, 0.0, 0.0, track_num, sector_num);
    ObjList_InsertObjToEnd(saucers, saucer, NULL);
  }

  set_saucers(0);

  return 0;
}

static int init()
{
  srand(time(NULL));

  camera.angle_dir = 0.0;
  camera.angle_up  = 20.0;
  camera.distance  = 5.0;

  make_saucers();

  gl_camera();
  gl_init();

  return 0;
}

static void gl_display()
{
  Saucer saucer;
  ObjListData obj;

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glLoadIdentity();

  gl_camera();
  gluLookAt(camera.position.x, camera.position.y, camera.position.z,
	    camera.direction.x, camera.direction.y, camera.direction.z,
	    camera.updir.x, camera.updir.y, camera.updir.z);
  gl_light();

  for (obj = ObjList_GetStart(saucers);
       !ObjList_IsEndEdge(saucers, obj);
       obj = ObjListData_GetNext(obj)) {
    saucer = (Saucer)ObjListData_GetObj(obj);

    glPushMatrix();
    Saucer_MakePolygon(saucer);
    glPopMatrix();
  }

  glutSwapBuffers();
}

static void gl_keyboard(unsigned char key, int x, int y)
{
  Saucer saucer;
  Block block;
  ObjListData obj;
  int n, t, s;

  switch (key) {
  case '\033':
  case 'q':
    exit(0);
  case 'i': camera.angle_up  += 0.5;  break;
  case 'k': camera.angle_up  -= 0.5;  break;
  case 'j': camera.angle_dir += 1.0;  break;
  case 'l': camera.angle_dir -= 1.0;  break;
  case 'o': camera.distance  += 0.03; break;
  case 'u': camera.distance  -= 0.03; break;

  case 't':
    for (obj = ObjList_GetStart(saucers);
	 !ObjList_IsEndEdge(saucers, obj);
	 obj = ObjListData_GetNext(obj)) {
      saucer = (Saucer)ObjListData_GetObj(obj);
      Saucer_AddTrackNum(saucer, 1);
    }
    track_num++;
    break;
  case 'g':
    for (obj = ObjList_GetStart(saucers);
	 !ObjList_IsEndEdge(saucers, obj);
	 obj = ObjListData_GetNext(obj)) {
      saucer = (Saucer)ObjListData_GetObj(obj);
      Saucer_AddSectorNum(saucer, 1);
    }
    sector_num++;
    break;

  case 'b':
    saucer = Saucer_Create(0.0, 0.0, 0.0, track_num, sector_num);
    ObjList_InsertObjToEnd(saucers, saucer, NULL);
    set_saucers(0);
    break;

  case 'r':
  case 'e':
    n = ObjList_GetLength(saucers);
    n = rand() % n;

    for (obj = ObjList_GetStart(saucers);
	 !ObjList_IsEndEdge(saucers, obj);
	 obj = ObjListData_GetNext(obj)) {
      if (n == 0)
	break;
      n--;
    }
    saucer = (Saucer)ObjListData_GetObj(obj);

    t = rand() % Saucer_GetTrackNum(saucer);
    s = rand() % Saucer_GetSectorNum(saucer);

    block = Saucer_GetBlock(saucer, t, s);

    switch (key) {
    case 'r':
      Block_AddCanon(block, Canon_Create(block), BLOCK_STATUS_ATTACK);
      break;
    case 'e':
      Block_AddCanon(block, Canon_Create(block), BLOCK_STATUS_TRY);
      break;
    default:
      break;
    }
    break;

  case 's':
    set_saucers(1);
    break;

  default:
    break;
  }
}

static void gl_special_keyboard(int key, int x, int y)
{
  switch (key) {
  case GLUT_KEY_UP:    gl_keyboard('i', x, y); return;
  case GLUT_KEY_DOWN:  gl_keyboard('k', x, y); return;
  case GLUT_KEY_LEFT:  gl_keyboard('j', x, y); return;
  case GLUT_KEY_RIGHT: gl_keyboard('l', x, y); return;
  default: break;
  }
}

static void gl_reshape(int width, int height)
{
  glViewport(0, 0, width, height);

  glMatrixMode(GL_PROJECTION);

  glLoadIdentity();

  gluPerspective(30.0, (double)width / (double)height, 1.0, 100.0);

  glMatrixMode(GL_MODELVIEW);
}

static int socket_proc();

static void gl_timeout(int value)
{
  Saucer saucer;
  ObjListData obj;

  socket_proc();

  for (obj = ObjList_GetStart(saucers);
       !ObjList_IsEndEdge(saucers, obj);
       obj = ObjListData_GetNext(obj)) {
    saucer = (Saucer)ObjListData_GetObj(obj);
    Saucer_Proc(saucer);
  }

  camera.angle_dir += 0.2;
  glutPostRedisplay();
  glutTimerFunc(33, gl_timeout, 0);
}

static int sockt = -1;

static int ObjList_GetIndex(ObjList list, Obj object)
{
  ObjListData obj;
  int i = 0;
  if (list == NULL) return -1;
  for (obj = ObjList_GetStart(list);
       !ObjList_IsEndEdge(list, obj);
       obj = ObjListData_GetNext(obj)) {
    if (ObjListData_GetObj(obj) == object)
      return i;
    i++;
  }
  return -1;
}

static Obj ObjList_SearchIndex(ObjList list, int index)
{
  ObjListData obj;
  if (list == NULL) return NULL;
  for (obj = ObjList_GetStart(list);
       !ObjList_IsEndEdge(list, obj);
       obj = ObjListData_GetNext(obj)) {
    if (index == 0)
      return ObjListData_GetObj(obj);
    index--;
  }
  return NULL;
}

static int socket_proc()
{
  Saucer saucer;
  Block block;
  User user;
  Genre genre;
  Question question;
  ObjListData obj;
  int t, s;
  char **argv;

  if (sockt < 0)
    return -1;

  argv = socket_readline(sockt, 0);
  if ((argv == NULL) || (argv[0] == NULL))
    return -1;

  if (!strcmp(argv[0], "USER")) {
    if ((argv[1] == NULL) || (argv[2] == NULL))
      return -1;
    if (User_ListSearch(users, argv[1]))
      return -1;

    if (ObjList_GetLength(users) < ObjList_GetLength(saucers)) {
      saucer = (Saucer)ObjList_SearchIndex(saucers, ObjList_GetLength(users));
    } else {
      saucer = Saucer_Create(0.0, 0.0, 0.0, track_num, sector_num);
      ObjList_InsertObjToEnd(saucers, saucer, NULL);
    }

    user = User_Create(argv[1], argv[2], saucer);
    ObjList_InsertObjToEnd(users, user, (ObjDestructor)User_Destroy);

    set_saucers(0);
  }

  if (!strcmp(argv[0], "QUESTION")) {
    if ((argv[1] == NULL) || (argv[2] == NULL) || (argv[3] == NULL))
      return -1;

    genre = Genre_ListSearch(genres, argv[2]);
    if (genre == NULL) {
      genre = Genre_Create(argv[2]);
      ObjList_InsertObjToEnd(genres, genre, (ObjDestructor)Genre_Destroy);
      while (ObjList_GetLength(genres) > track_num) {
	for (obj = ObjList_GetStart(saucers);
	     !ObjList_IsEndEdge(saucers, obj);
	     obj = ObjListData_GetNext(obj)) {
	  saucer = (Saucer)ObjListData_GetObj(obj);
	  Saucer_AddTrackNum(saucer, 1);
	}
	track_num++;
      }
    }

    if (Question_ListSearch(Genre_GetQuestions(genre), argv[1]))
      return -1;

    question = Question_Create(argv[1], atoi(argv[3]), genre);
    ObjList_InsertObjToEnd(Genre_GetQuestions(genre), question,
			   (ObjDestructor)Question_Destroy);

    while (ObjList_GetLength(Genre_GetQuestions(genre)) > sector_num) {
      for (obj = ObjList_GetStart(saucers);
	   !ObjList_IsEndEdge(saucers, obj);
	   obj = ObjListData_GetNext(obj)) {
	saucer = (Saucer)ObjListData_GetObj(obj);
	Saucer_AddSectorNum(saucer, 1);
      }
      sector_num++;
    }
  }

  if (!strcmp(argv[0], "REPLY") || !strcmp(argv[0], "CLEAR")) {
    if ((argv[1] == NULL) || (argv[2] == NULL) || (argv[3] == NULL))
      return -1;

    user = User_ListSearch(users, argv[1]);
    if (user == NULL)
      return -1;
    question = Question_ListSearchAll(genres, argv[2]);
    if (question == NULL)
      return -1;
    genre = Question_GetGenre(question);
    saucer = User_GetSaucer(user);

    t = ObjList_GetIndex(genres, (ObjListData)genre);
    s = ObjList_GetIndex(Genre_GetQuestions(genre), (ObjListData)question);
    if ((t < 0) || (s < 0))
      return -1;

    block = Saucer_GetBlock(saucer, t, s);

    if (!strcmp(argv[0], "REPLY"))
      Block_AddCanon(block, Canon_Create(block), BLOCK_STATUS_TRY);
    if (!strcmp(argv[0], "CLEAR"))
      Block_AddCanon(block, Canon_Create(block), BLOCK_STATUS_ATTACK);
  }

  return 0;
}

static char *server_name = NULL;
static int port = 12345;
static int server_mode = 0;
static int stand_alone = 0;

static Argument args[] = {
  { "-host"      , ARGUMENT_TYPE_STRING  , &server_name },
  { "-h"         , ARGUMENT_TYPE_STRING  , &server_name },
  { "-port"      , ARGUMENT_TYPE_INTEGER , &port },
  { "-p"         , ARGUMENT_TYPE_INTEGER , &port },
  { "-server"    , ARGUMENT_TYPE_FLAG_ON , &server_mode },
  { "-s"         , ARGUMENT_TYPE_FLAG_ON , &server_mode },
  { "-alone"     , ARGUMENT_TYPE_FLAG_ON , &stand_alone },
  { "-saucer-num", ARGUMENT_TYPE_INTEGER , &saucer_num },
  { "-track-num" , ARGUMENT_TYPE_INTEGER , &track_num },
  { "-sector-num", ARGUMENT_TYPE_INTEGER , &sector_num },
  { NULL     , ARGUMENT_TYPE_NONE    , NULL },
};

int main(int argc, char *argv[])
{
  argument_read(&argc, argv, args);

  if (!stand_alone) {
    if (server_mode) {
      sockt = socket_open_server(port);
    } else if (server_name) {
      sockt = socket_open_client(server_name, port);
    }
  }

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
  glutInitWindowSize(800, 600);
  glutCreateWindow(argv[0]);

  glutDisplayFunc(gl_display);
  glutKeyboardFunc(gl_keyboard);
  glutSpecialFunc(gl_special_keyboard);
  glutReshapeFunc(gl_reshape);
  glutTimerFunc(33, gl_timeout, 0);

  init();

  glutMainLoop();
  return 0;
}
